/*
 * Copyright 2004, Red Hat, Inc. All Rights Reserved
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * on the rights to use, copy, modify, merge, publish, distribute, sub
 * license, and/or sell copies of the Software, and to permit persons to whom
 * the Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice (including the next
 * paragraph) shall be included in all copies or substantial portions of the
 * Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT.  IN NO EVENT SHALL
 * RED HAT OR ITS SUPPLIERS BE LIABLE FOR ANY CLAIM,
 * DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
 * OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE
 * USE OR OTHER DEALINGS IN THE SOFTWARE.
 *
 * Author:
 *    Owen Taylor <otaylor@redhat.com>
 */
#include <errno.h>
#include <fcntl.h>
#include <signal.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include <netinet/in.h>

#include <sys/ioctl.h>
#include <sys/poll.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <sys/types.h>

#include "texprof-lib.h"
#include "utils.h"

/****************************************************************/

/* Global state */

static unsigned short port = 0;
static int socketFD = -1;

static Profile *profile;
xbool needDisplay = TRUE;
xbool inUpdate = FALSE;
struct timeval lastUpdateTime;

static int connectionFD = -1;
static char connectionOutputBuffer[PROFILE_MAX_LEN];
static size_t connectionOutputLen;
static size_t connectionOutputPos;
static char connectionInputBuffer[PROFILE_MAX_LEN];
static size_t connectionInputLen;

static char *connectionError;

#define UPDATE_DELAY 1000


/****************************************************************
 * Utilities
 ****************************************************************/

static char *programName;

void
fatal (const char *msg, ...)
{
    va_list vap;
    
    if (programName)
	fprintf(stderr, "%s: ", programName);
    
    va_start(vap, msg);
    vfprintf(stderr, msg, vap);
    va_end(vap);
    
    fprintf(stderr, "\n");
    
    exit(1);
}

static void
fatal_errno (const char *msg, ...)
{
    va_list vap;
    int err = errno;
    
    if (programName)
	fprintf(stderr, "%s: ", programName);
    
    va_start(vap, msg);
    vfprintf(stderr, msg, vap);
    va_end (vap);

    fprintf(stderr, ":  %s", strerror(err));
    
    fprintf(stderr, "\n");
    
    exit(1);
}

void *
xmalloc (size_t bytes)
{
    void *retval = malloc(bytes);
    if (!retval)
	fatal("Out of memory");
    
    return retval;
}

void *
xcalloc (size_t nmemb, size_t bytes)
{
    void *retval = calloc(nmemb, bytes);
    if (!retval)
	fatal("Out of memory");
    
    return retval;
}

void *
xrealloc (void  *ptr,
	  size_t bytes)
{
    void *retval = realloc(ptr, bytes);
    if (!retval)
	fatal("Out of memory");
    
    return retval;
}

void *
xstrdup (const char *str)
{
    if (str) {
	size_t len = strlen(str);
	void *retval = xmalloc(len + 1);
	memcpy(retval, str, len + 1);
	
	return retval;
    } else
	return NULL;
}

static void
setProgramName(const char *argv0)
{
    char *p;
    
    programName = xstrdup(argv0);
    p = strrchr(programName, '/');
    if (p)
	memmove(programName, p + 1, programName + strlen(programName) - p);
}

static double
timeDifference(struct timeval *current,
	       struct timeval *base)
{
    return ((current->tv_sec - base->tv_sec) * 1000.0 +
	    (current->tv_usec - base->tv_usec) / 1000.0);
}

/****************************************************************
 * Femto-curses: A few routines for full-screen terminal output
 ****************************************************************/

static xbool have_winsize = 0;
static int rows = 24, columns = 80;

typedef struct {
  int row;
  int column;
    xbool reverse;
} ScreenState;

static void
onSIGWINCH(int signum)
{
    have_winsize = 0;
    needDisplay = TRUE;
}

static void
setupSIGWINCH(void)
{
    struct sigaction sa;

    sa.sa_handler = onSIGWINCH;
    sigemptyset(&sa.sa_mask);
    sa.sa_flags = 0;

    sigaction(SIGWINCH, &sa, NULL);
}

/* A few very basic VT100 escape sequences */
#define CLEAR_SCREEN "\E[J"
#define HOME_CURSOR "\E[H"
#define CLEAR_LINE "\E[K"
#define REVERSE_VIDEO "\E[7m"
#define NORMAL_VIDEO "\E[m"

static void
initScreen()
{
    setupSIGWINCH();
    printf(CLEAR_SCREEN);
}

static void
startScreen (ScreenState *state)
{
    /* Update current terminal size, if necessary */
    if (!have_winsize) {
	struct winsize ws;
	
	have_winsize = 1;
	
	if (ioctl(0, TIOCGWINSZ, &ws) != -1) {
	    rows = ws.ws_row;
	    columns = ws.ws_col;
	}
    }
    
    state->row = 0;
    state->column = 0;
    state->reverse = FALSE;
    printf(HOME_CURSOR);    
}

static void
reverseVideo (ScreenState *state)
{
    printf(REVERSE_VIDEO);
    state->reverse = TRUE;
}

static void
normalVideo (ScreenState *state)
{
    printf(NORMAL_VIDEO);
    state->reverse = FALSE;
}

static void
printfRow (ScreenState *state, const char *format, ...)
{
    va_list vap;
    int i;
    xbool last_row;
    
    va_start(vap, format);
    state->column += vprintf(format, vap);
    va_end(vap);

    last_row = state->row == rows - 1;

    if (state->reverse) {
	for (i = state->column; i < (last_row ? columns - 1 : columns); i++)
	    putc(' ', stdout);
    } else {
	if (last_row)
	    printf(CLEAR_LINE);
	else
	    printf(CLEAR_LINE "\n");
    }

    state->row++;
    state->column = 0;
}

static void
endScreen (ScreenState *state)
{
    int i;
    
    for (i = state->row; i < rows - 1; i++)
	printf(CLEAR_LINE "\n");
    printf(CLEAR_LINE);
    fflush(stdout);
}

/*****************************************************************
 * Callbacks
 *****************************************************************/

static void
outputLine (Profile    *profile,
	    const char *line)
{
    if (connectionOutputLen > 0)
	fatal("outputLine called while another line is still be written");

    connectionOutputLen = strlen(line) + 2;
    memcpy(connectionOutputBuffer, line, connectionOutputLen - 2);
    connectionOutputBuffer[connectionOutputLen - 2] = '\r';
    connectionOutputBuffer[connectionOutputLen - 1] = '\n';
    
}

static void
beginUpdate (Profile *profile)
{
}

static void
endUpdate (Profile *profile)
{
    needDisplay = TRUE;
    inUpdate = FALSE;
}

static const ProfileFuncs profileFuncs =
{
    outputLine,
    beginUpdate,
    endUpdate,
};

/*****************************************************************/

static void
closeConnection(const char *err)
{
  if (connectionError) {
	free(connectionError);
	connectionError = NULL;
    }

    connectionError = xstrdup(err);

    profileDestroy(profile);
    profile = NULL;
    
    close(connectionFD);
    connectionFD = -1;
    
    connectionOutputPos = 0;
    connectionOutputLen = 0;
    connectionInputLen = 0;

    needDisplay = TRUE;
}

static void
createSocket(void)
{
    struct sockaddr_in addr;
    socklen_t addrlen = sizeof(addr);

    socketFD = socket(PF_INET, SOCK_STREAM, 0);
    if (socketFD == -1)
	fatal_errno("createSocket: socket");

    if (fcntl(socketFD, F_SETFL, O_NONBLOCK) == -1)
	fatal_errno("createSocket: fcntl");

    addr.sin_family = AF_INET;
    addr.sin_port = htons(port);
    addr.sin_addr.s_addr = htonl(INADDR_LOOPBACK);

    if (bind(socketFD, (struct sockaddr *)&addr, addrlen) == -1)
	fatal_errno("createSocket: bind");

    if (listen(socketFD, 16 /* backlog */) == -1)
	fatal_errno("createSocket: bind");

    if (port == 0) {
	if (getsockname(socketFD, (struct sockaddr *)&addr, &addrlen) == -1)
	    fatal_errno("createSocket: getsockname");

	port = ntohs(addr.sin_port);
    }
}

static void
doConnectionUpdate(struct timeval *current_time)
{
    if (!current_time)
	gettimeofday(&lastUpdateTime, NULL);
    else
	lastUpdateTime = *current_time;
    
    if (profile) {
	profileUpdate(profile);
	inUpdate = TRUE;
    }
}

static void
acceptConnection(void)
{
    struct sockaddr_in addr;
    socklen_t addrlen = sizeof(addr);
    int fd;

    addrlen = sizeof(struct sockaddr_in);

 again:
    fd = accept(socketFD, (struct sockaddr *)&addr, &addrlen);
    if (fd == -1) {
	if (errno == EINTR)
	    goto again;
	if (errno == EAGAIN)
	    return;
    
	fatal_errno("acceptConnection: accept");
    }

    if (profile) {	/* Only one connection at once, reject immediately */
	close(fd);
	return;
    }
    
    if (fcntl(fd, F_SETFL, O_NONBLOCK) == -1)
	fatal_errno("acceptConnection: fcntl");

    connectionFD = fd;
    profile = profileNew(programName, &profileFuncs, sizeof(profileFuncs), NULL);
    doConnectionUpdate(NULL);

    needDisplay = TRUE;
    
    return;
}

static void
writeConnectionOutput(void)
{
    while (connectionOutputPos < connectionOutputLen) {
	/* Use send(2) rather than write(2) so we can avoid SIGPIPE */
	int count = send(connectionFD,
			 connectionOutputBuffer + connectionOutputPos,
			 connectionOutputLen - connectionOutputPos,
			 MSG_NOSIGNAL);
	if (count < 0) {
	    if (errno == EINTR)	/* Interrupted by signal */
		continue;
	    else if (errno == EAGAIN) /* Would block */
		break;
	    else if (errno == EPIPE) { /* EOF */
		closeConnection(NULL);
		return;
	    } else
		fatal_errno("writeConnectionOutput: send");
	} else
	    connectionOutputPos += count;
    }

    if (connectionOutputPos == connectionOutputLen) {
	connectionOutputPos = 0;
	connectionOutputLen = 0;
    }
}

static xbool
processConnectionLine(const char *line)
{
    const char *err = profileFeedInputLine(profile, line);
    if (err) {
	char *tmp = xmalloc(strlen(err) + strlen(line) + 3);
	strcpy(tmp, err);
	strcat(tmp, ": ");
	strcat(tmp, line);
	closeConnection(tmp);
	free(tmp);
	return FALSE;
    } else
	return TRUE;
}

static void
readConnectionInput(void)
{
    char *startpos = NULL, *endpos = NULL;
    size_t remaining;

    while (connectionInputLen < PROFILE_MAX_LEN) {
	int count = read(connectionFD,
			 connectionInputBuffer + connectionInputLen,
			 PROFILE_MAX_LEN - connectionInputLen);
	
	if (count < 0) {
	    if (errno == EINTR)	/* Interrupted by signal */
		continue;
	    else if (errno == EAGAIN) /* Would block */
		break;
	    else if (errno == ECONNRESET) {
		closeConnection(NULL);
		return;
	    } else
		fatal_errno("readConnectionInput: read");
	} else if (count == 0) { /* EOF */
	    closeConnection(NULL);
	    return;
	} else
	    connectionInputLen += count;
    }

    startpos = connectionInputBuffer;

    while (profile && startpos < connectionInputBuffer + connectionInputLen) {
	endpos = memchr(startpos, '\r', connectionInputLen);
	if (endpos && endpos < connectionInputBuffer + connectionInputLen - 1) {
	    if (*(endpos + 1) != '\n') {
		
		closeConnection("Client lines not properly separated by \\r\\n\n");
		return;
	    }

	    *endpos = '\0';
	    
	    if (!processConnectionLine(startpos))
		return;
	    
	    startpos = endpos + 2;
	} else
	    break;
    }

    remaining = connectionInputBuffer + connectionInputLen - startpos;

    /* If have a full buffer remaining, then the application has
     * sent a longer line than permitted in the protocol
     */
    if (remaining == PROFILE_MAX_LEN)
      closeConnection("Overlong command from client");
    else {
      memmove(connectionInputBuffer, startpos, remaining);
      connectionInputLen = remaining;
    }
}

static void
handleNetworkIO(void)
{
    enum { SOCKET, CONNECTION };
    
    struct pollfd fds[2];
    xbool have_output;
    int count = 0;
    struct timeval current_time;
    double delay;

    fds[SOCKET].fd = socketFD;
    fds[SOCKET].events = POLLIN;
    fds[SOCKET].revents = 0;
    count++;

    if (connectionFD != -1) {
	have_output = connectionOutputLen > 0;
	
	fds[CONNECTION].fd = connectionFD;
	fds[CONNECTION].events = have_output ? POLLOUT : POLLIN;
	fds[CONNECTION].revents = 0;
	count++;
    } else {
	have_output = FALSE;
    }

    gettimeofday (&current_time, NULL);
    delay = UPDATE_DELAY - timeDifference (&lastUpdateTime, &current_time);
    if (delay < 0)
	delay = 0;
    else if (delay > UPDATE_DELAY)
	delay = UPDATE_DELAY;
    
    poll(fds, count, (int)delay);
    
    if (fds[SOCKET].events & POLLIN) { /* New connection */
	acceptConnection();
    }

    if (connectionFD != -1) {
	if (have_output) {
	    if (fds[CONNECTION].revents & (POLLOUT))
		writeConnectionOutput();
	} else {
	    if (fds[CONNECTION].revents & (POLLIN | POLLHUP))
		readConnectionInput();
	}
    }

    if (!inUpdate) {
	double difference;
	
	gettimeofday(&current_time, NULL);
	difference = timeDifference(&current_time, &lastUpdateTime);
	if (difference < 0 || difference > UPDATE_DELAY)
	    doConnectionUpdate(&current_time);
    }
}

/* Turn a size in bytes into a 5 character or less string
 * (for sizes up to 9999M) */
static char *
sizeString(unsigned bytes)
{
    char buf[256];

    if (bytes >= 3 * 1024 * 1024) {
	snprintf(buf, sizeof(buf), "%.0fM", bytes / (1024. * 1024));
    } else if (bytes >= 0.3 * 1024 * 1024) {
	snprintf(buf, sizeof(buf), "%.1fM", bytes / (1024. * 1024));
    } else if (bytes >= 0.1 * 1024 * 1024) {
	snprintf(buf, sizeof(buf), "%.2fM", bytes / (1024. * 1024));
    } else if (bytes >= 10 * 1024) {
	snprintf(buf, sizeof(buf), "%.0fK", bytes / (1024.));
    } else if (bytes >= 1024) {
	snprintf(buf, sizeof(buf), "%.1fK", bytes / (1024.));
    } else
	snprintf(buf, sizeof(buf), "%db", bytes);

    return xstrdup(buf);
}

/* Turn a time in seconds into a 5 character or less string
 * (for times up to 9999h) */
static char *
timeString(double t)
{
    char buf[256];

#define SECOND 1000.
#define MINUTE (60. * SECOND)
#define HOUR (60. * MINUTE)
    if (t > 10 * HOUR)
	snprintf(buf, sizeof(buf), " %.0fh", t / HOUR);
    else if (t > HOUR)
	snprintf(buf, sizeof(buf), " %.1fh", t / HOUR);
    else if (t > 10. * MINUTE)
	snprintf(buf, sizeof(buf), " %.0fm", t / MINUTE);
    else if (t > MINUTE)
	snprintf(buf, sizeof(buf), " %.1fm", t / MINUTE);
    else if (t > 10. * SECOND)
	snprintf(buf, sizeof(buf), " %.0fs", t / SECOND);
    else if (t > 1. * SECOND)
	snprintf(buf, sizeof(buf), " %.1fs", t / SECOND);
    else
	snprintf(buf, sizeof(buf), " %.2fs", t / SECOND);
#undef HOUR
#undef MINUTE
#undef SECOND

    return xstrdup(buf);
}

static void
printCurrentState(void)
{
    ScreenState state;

    if (!needDisplay)
	return;

    needDisplay = FALSE;

    /* Print current state
     */
    startScreen(&state);

    if (profile) {
	ProfileHeap * const * heaps;
	ProfileTexture * const * textures;
	int n_heaps, n_textures;
	struct timeval update_timestamp;
	int i;
	int max_heap_len = 0;

	profileGetUpdateTimestamp(profile, &update_timestamp);

	profileGetHeaps(profile, &heaps, &n_heaps);

	for (i = 0; i < n_heaps; i++) {
	    const char *name = "<unknown>";

	    if (heaps[i]->flags & PROFILE_HEAP_NAME)
		name = heaps[i]->name;

	    int len = strlen(name);
	    if (len > max_heap_len)
		max_heap_len = len;
	}
	
	for (i = 0; i < n_heaps; i++) {
	    const char *name = "<unknown>";
	    unsigned size = 0;
	    char *size_string;

	    if (heaps[i]->flags & PROFILE_HEAP_NAME)
		name = heaps[i]->name;

	    if (heaps[i]->flags & PROFILE_HEAP_SIZE)
		size = heaps[i]->size;

	    if (size)
		size_string = sizeString(size);
	    else
		size_string = xstrdup ("");

	    printfRow(&state, "%s %*s %5s",
		      i == 0 ? "Heaps:" : "      ",
		      max_heap_len, name, size_string);

	    free(size_string);
	}

	printfRow(&state, "");

	reverseVideo(&state);
	printfRow(&state, " width height main_size %*s heap_age heap_size",
		   max_heap_len, "heap");
	normalVideo(&state);

	profileGetTextures(profile, &textures, &n_textures);
	for (i = 0; i < n_textures && state.row < rows ; i++) {
	    ProfileHeap *heap = NULL;
	    const char *heap_name = "";
	    unsigned width = 0, height = 0;
	    char *heap_age_string, *main_size_string, *heap_size_string;

	    if (textures[i]->flags & PROFILE_TEXTURE_WIDTH)
		width = textures[i]->width;
	    if (textures[i]->flags & PROFILE_TEXTURE_HEIGHT)
		height = textures[i]->height;
	    if (textures[i]->flags & PROFILE_TEXTURE_HEAP)
		heap = textures[i]->heap;
	    if (heap && (heap->flags & PROFILE_HEAP_NAME))
		heap_name = heap->name;

	    if ((textures[i]->flags & PROFILE_TEXTURE_MAIN_SIZE) &&
		textures[i]->mainSize)
		main_size_string = sizeString(textures[i]->mainSize);
	    else
		main_size_string = xstrdup("");
	    
	    if (textures[i]->flags & PROFILE_TEXTURE_HEAP_TS) {
		double age = timeDifference(&update_timestamp, &textures[i]->heapTS);
		heap_age_string = timeString(age);
	    } else {
		heap_age_string = xstrdup("");
	    }

	    if ((textures[i]->flags & PROFILE_TEXTURE_HEAP_SIZE) &&
		textures[i]->heapSize)
		heap_size_string = sizeString(textures[i]->heapSize);
	    else
		heap_size_string = xstrdup("");
	    
	    printfRow(&state, "%6d %6d %9s %*s %8s %9s",
		      width, height, main_size_string,
		      max_heap_len, heap_name,
		      heap_age_string, heap_size_string);

	    free(main_size_string);
	    free(heap_age_string);
	    free(heap_size_string);
	}
    } else {
	if (connectionError) {
	    printfRow(&state, "Last client disconnected because of an error:", port);
	    printfRow(&state, "    %s", connectionError);
	}
	printfRow(&state, "Waiting for a connection on port %d", port);
    }
    
    endScreen(&state);
}

int
main(int argc, char **argv)
{
    setProgramName(argv[0]);
    
    if (argc != 1 && argc != 2) {
	fprintf(stderr, "Usage: %s [PORT]\n", programName);
    }

    if (argc == 2) {
	char *end;
	long l;
	
	port = l = strtol(argv[1], &end, 0);
	if (argv[1][0] == '\0' || *end != '\0' || l != port) {
	    fatal("Invalid port '%s'", argv[1]);
	}
    }

    createSocket();
    initScreen();
    
    while (1) {
	printCurrentState();
	handleNetworkIO();
    }
    
    return 0;
}
